#  - conf.py - 
# This file is part of mp3togo

# Convert audio files to play on a mp3 player
# Manage program options
#
# (c) Simeon Veldstra 2004, 2006  <reallifesim@gmail.com>
#
# This software is free.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You may redistribute this program under the terms of the
# GNU General Public Licence version 2
# Available in this package or at http://www.fsf.org

# requires python-pyvorbis and python-id3 and lame
# and mpg321

import sys, os

import mp3togo
import mp3togo.options as options


class Fail(options.Fail):
	pass

class Error(Exception):
	pass 
	
class ErrorUnknownFileType(Error):
	pass

class ErrorNoFile(Error):
	pass

class ErrorNoCache(Error):
	pass

class ErrorBadFormat(Error):
	pass

class ErrorNoXMMSControl(Error):
	pass

class ErrorXMMSNotRunning(Error):
	pass

class ErrorNoDCOP(Error):
	pass

class ErrorAmarokNotRunning(Error):
	pass

class ErrorUnlocked(Error):
	pass

class TaskNotReadyError(Error):
	pass

class ErrorClusterProtocol(Error):
	pass



class Options(options.Options):
	"""Subclass options.Options and fill in application specific
	information."""
	def __init__(self, argv=None, conffile=None, readconf=True):
		options.Options.__init__(self, '~/mp3togo')

		# Options to add:
		# ((name, short option, long option, default value, hook, help text), ...)
		self._set += [
 ('brwarning', 't', 'file-tag', '.2go', None, 
 '''A string appended to the name of the file to indicate it has been recoded.'''),
 ('tempdir', 'w', 'work-dir', '/tmp', self._absfile, 
 '''The path to store temporary files. This could need several hundred megabytes free.'''),
 ('treedepth', 'd', 'tree-depth', 1, None, 
 '''The number of directory levels to preserve in the output tree. Use 0 to put all output files directly into the target directory. Use 2 to create a directory for the Artist and then the Album (Assuming your music collection is organized in directories by artist then album).'''),
 ('treestructure', '', 'format', '', None, 
 '''A string specifying the format for the output files. The format string can contain the following escapes:<li>%a - Artist<li>%l - Album<li>%t - Title<li>%y - Year<li>%g - Genre<li>%% - Literal '%'<br>Overrides --tree-depth.'''),
 ('maxunits', 'm', 'max-size', '0', self._units, 
 '''The disk space available to use in bytes. Append 'M' for megabytes, 'G' for gigabytes or 'K' for kilobytes. Use 0 to use all available space on the filesystem.'''),
 ('maxsize', '', '', 0L, None, ''),
 ('maxtempunits', '', 'max-temp-size', '0', self._units, 
 '''The disk space available to use for temporary files in bytes. Append 'M' for megabytes, 'G' for gigabytes or 'K' for kilobytes. Use 0 to use all available space on the filesystem.'''),
 ('maxtempsize', '', '', 0L, None, ''),
 ('force', 'F', 'force', False, None, 
 '''Overwrite files if they already exist.'''),
 ('encoder', 'C', 'encoder', 'lame', None, 
 '''The encoder to use to create the output files. Options are wav, lame or oggenc.'''),
 ('encopts', 'E', 'encoder-options', '', None, 
 '''Compression options for the encoder.'''),
 ('compfactor', 'z', 'compression-factor', 16.0, None, 
 '''If you change the lame options, you must change this factor to match the compression rate of the new options. This is used to estimate completed file size.'''),
 ('makecache', '', 'make-cache', '', self._absfile, '''Make an output file cache at the given path.'''),
 ('cachesize', '', '', 0L, None, ''),
 ('cacheunits', '', 'cache-size', '0', self._units, '''The size for the new cache.'''),
 ('usecache', '', 'use-cache', '', self._absfile, '''Use the cache stored at the given path.'''),
 ('help', 'h', 'help', False, None, '''Print this message.'''),
 ('playlist', 'p', 'playlist', '', None, 
 '''The playlist to convert. Playlists can be simple lists of file paths, one per line, or .m3u files or the native XMMS playlist file format. Playlists can be also listed on the command line as free arguments after all of the options if they have a recognizable extension. Currently .m3u and .pls are recognized.'''),
 ('readxmms', 'x', 'import-xmms', False, None, 
 '''Get playlist from running instance of XMMS. (You must have the python-xmms module installed)'''),
 ('readamarok', '', 'import-amarok', False, None,
 '''Get playlist from running instance of Amarok.'''),
 ('playerdir', 'o', 'output-dir', os.curdir, self._absfile, 
 '''Where to write the output files.'''),
 ('index', '', 'index', False, None, 
 '''Create an index file when in wav output mode.'''),
 ('nonormal', '', 'no-normalize', False, None, 
 '''Don't normalize the wav file before encoding.'''),
 ('cluster', '', 'cluster', '', None, 
 '''Manage a cluster of worker machines. Cluster mode distributes tracks between multiple computers connected to a LAN. Provide a comma separated list of IP addresses or hostnames. The master machine must be able to log in to the slaves without a password and the slaves must have shared access to the filesystem. See the website for an example of how to set it up.<li>http://puddle.ca/mp3togo/'''),
 ('clusternolocal', '', 'cluster-no-local-node', False, None, 
 '''Only use the remote nodes for processing files.'''),
 ('clusterslave', '', 'cluster-slave', False, None, 
 '''Start a worker process for cluster mode. This option is used by the cluster master to start slave processes. Do not use.''')]

		# Override hook defined in Base class
		self._conffile = self._absfile 

		# Options to not save to the config file:
		self._nosave_options += []

		# Binaries to check for:
		self.bin['wav'] = True
		self.bin['lame'] = False
		self.bin['oggenc'] = False
		self.bin['mpg321'] = False
		self.bin['ogg123'] = False
		self.bin['flac'] = False
		self.bin['metaflac'] = False
		self.bin['normalize-audio'] = False
		self.bin['dcop'] = False

		self.version = mp3togo.version
		self._help_preamble = """
Synopsis:
 mp3togo [-p playlist] [-o output-dir] [options] [input files]

Description:
 mp3togo is a program for converting audio files of various
 formats into a common format suitable for use on a portable
 mp3 player or an mp3 CD. The files to convert can be
 specified in a playlist file (m3u, pls and plain text are
 supported) or given as arguments to the program. mp3togo
 will go through the list and run mpg321, ogg123, flac and
 lame to decode and reencode the files. The newly encoded
 files will be written to the destination directory. The
 software can retain as much of the subdirectory structure
 as desired.
"""

		# Check for Third party modules:
		self.mod['ogg.vorbis'] = False
		self.mod['ID3'] = False
		self.mod['xmms'] = False

		# Go ahead and read in the data:
		self.getconf(argv, conffile, readconf)



	# All hooks are called with the lock held
	# and must reference the ._d dict directly
	def _post_hook(self):
		#Set up default encoder options if not specified:
		if not self._d['encopts']:
			self.reset_encoder_options()

	def _absfile(self, name, value):
		self._d[name] = absfile(value)

	def _units(self, name, value):
		#Calculate raw bytes and set [max|temp|cache]size
		try:
			if value[-1] in ('m', 'M'):
				max = long(value[:-1]) * 1048576L
			elif value[-1] in ('g', 'G'):
				max = long(value[:-1]) * 1073741824L
			elif value[-1] in ('k', 'K'):
				max = long(value[:-1]) * 1024L
			else:
				max = long(value)
		except ValueError:
			raise Fail, "Bad %s option: %s " % (name, value)
		sizename = name.replace('units', 'size')
		self._d[sizename] = max    # 'maxsize'
		self._d[name] = value # 'maxunits'

	def _arg_files(self, name, value):
		l = []
		for f in value:
			f = absfile(f)
			if os.path.exists(f):
				l.append(f)
		self._d[name] = l

	def reset_encoder_options(self):
		"""Reset encoder options to defaults."""
		if self._d['encoder'] == 'lame':
			self._d['encopts'] = '--abr 96'
		elif self._d['encoder'] == 'oggenc':
			self._d['encopts'] = '-m 96 -M 225 -b 100'
		elif self._d['encoder'] == 'wav':
			self._d['encopts'] = ''

	def cleanfilename(self, name):
		"""Remove nasty characters from a filename"""
		name = name.replace('"', "'")
		name = name.replace(':', '.')
		name = name.replace("?", "")
		name = name.replace('*', '')
		name = name.replace('`', "'")
		name = name.replace('&', '+')
		name = name.replace(';', '.')
		name = name.replace('#', '-')
		if name[0] == '.':
			name = '_' + name
		return name

def absfile(name):
	name = name.strip()
	name = os.path.expanduser(name)
	if name:
		name = os.path.abspath(name)
	return name

def getfiletype(filename):
	"""Return the format of an audio file"""
	# Could use some more magic here
	ft = os.path.splitext(filename)[1][1:]
	if ft not in ('mp3', 'ogg', 'flac', 'wav'):
		raise ErrorUnknownFileType
	return ft

def checkfile(name):
	"""Verify the existence of an audio file"""
	name = name.replace('\n', '')
	if not os.path.exists(name):
		raise ErrorNoFile
	name = absfile(name)
	getfiletype(name) # Throws exception if unknown
	return name

def try_print(msg):
	"""Try to print a message to a nonblocking stdout"""
	for i in range(5):
		try:
			sys.stdout.write(msg)
			return
		except:
			time.sleep(0.01 ** (1.0 / i))

