# helpers.py: functions to interpret coder program output.
# mp3togo  Portable Music Manager
#
# Copyright 2005:  Simeon Veldstra <reallifesim@gmail.com> 
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# ***********************************************************
# The code to parse output from external programs is based on 
# the jack CD ripping program by Arne Zellentin.
#
# 'jack' is responsible for my full hard disks
# and has inspired me to produce this program.
# Thanks for everything Arne.


import os
import shutil

import mp3togo.conf as conf


class frames:
	def __init__(self):
		self.frames = 0
		self.tframes = 0

def parse_ogg123(buf):
	s = buf.split('\r')
	def sumtime(t):
		sum = float(t.split(':')[0]) * 60
		sum += float(t.split(':')[1])
		return sum
	for l in s:
		if l.startswith('Time') and l.endswith('% '):
			line = l.split(' ')
			return sumtime(line[1])/sumtime(line[4]) * 100.0
	return 0 

def parse_mpg321(buf, f=frames()):
	s = buf.split('\r')
	if len(s) >= 2:
		s = s[-2]
	else:
		s = s[-1]
	if not f.tframes:
		y0 = s.find('[')
		y1 = s.find(']')
		if y0 != -1 and y1 != -1:
			try:
				f.tframes = float(s[y0+1:y1])
			except ValueError:
				pass
	if f.tframes:
		try:
			r = float(s.split()[1])
		except ValueError:
			r = f.tframes
		if r < f.frames:
			f.tframes = f.frames = 0
			return 0
		f.frames = r
		return f.frames/f.tframes * 100.0
	return 0

def parse_oggenc(buf):
	# From "jack"
	s = buf.split('\r')
	if len(s) >= 2:
		s = s[-2]
	if len(s) == 1:
		s = s[0]
	y0 = s.find("[")
	y1 = s.find("%]")
	if y0 != -1 and y1 != -1:
		percent = float(s[y0 + 1:y1])
	else:
		percent = 0
	return percent

def parse_lame(buf):
	# originaly from "jack"
	s = buf.split('\r')
	if len(s) >= 2: 
		s = s[-1] or s[-2]
	if len(s) == 1: s=s[0]
	if s.find("%") >= 0:       # status reporting starts here
		y = s.split("/")
		y1 = y[1].split("(")[0]
		percent = float(y[0]) / float(y1) * 100.0
	elif s.find("Frame:") >= 0:    # older versions, like 3.13
		y = s.split("/")
		y0 = y[0].split("[")[-1]
		y1 = y[1].split("]")[0]
		percent = float(y0) / float(y1) * 100.0
	else:
		percent = 0
	return percent

def parse_flac_enc(buf):
	# From "jack" 
	s = buf.split('\r')    
	if len (s) >= 2: s = s[-2]
	if len (s) == 1: s = s[0]       
	y0 = s.find(": ")             
	y1 = s.find("%")
	if y0 != -1 and y1 != -1: 
	    return float(s[y0 + 1:y1])
	else:                   
	    return 0 

def parse_flac_dec(buf):
	# I wrote this one myself
	buf =  buf.split('\r')
	buf =  buf[-1]
	buf = buf.replace('%', '')
	try:
		return int(buf.split()[1])
	except IndexError:
		return 0

### wav file "encoder/decoder"
def recode_wav(input, output, args=None):
	"""Factory returns function to copy wav file

	Returned function takes no args and copies
	input to output. Throws OSError on error."""
	return lambda: shutil.copyfile(input, output)

## Command line escape sequences:
#
#  escape:      for:
#  ---------------------
#  %%       	%
#  %o			output filename
#  %i 			input filename
#  %a			args

helpers = {
 'oggenc' : {'parser': parse_oggenc,
             'cmd': 'oggenc %a -o %o %i',
			 'type': 'ogg',
			 'action': 'encode'},
 'lame' :   {'parser': parse_lame,
             'cmd': 'lame --nohist %a %i %o',
		     'type': 'mp3',
		     'action': 'encode'},
 'flac_enc':{'parser': parse_flac_enc,
		     'cmd': 'flac ???????',
		     'type': 'flac',
		     'action': 'encode'},
 'wav_enc': {'parser': None,
 			 'cmd': recode_wav,
			 'type': 'wav',
			 'action': 'encode'},
 'ogg123' : {'parser': parse_ogg123,
             'cmd': 'ogg123 -d wav -f %o %i',
			 'type': 'ogg',
			 'action': 'decode'},
 'mpg321' : {'parser': parse_mpg321,
             'cmd': 'mpg321 -v -w %o %i',
			 'type': 'mp3',
			 'action': 'decode'},
 'flac_dec':{'parser': parse_flac_dec,
		     'cmd': 'flac --decode -F -o %o %i',
		     'type': 'flac',
		     'action': 'decode'},
 'wav_dec': {'parser': None,
 			 'cmd': recode_wav,
			 'type': 'wav',
			 'action': 'decode'},
 }


def find_helper(type, action):
	"""Return the name of the helper that performs action on type
	
	Raises KeyError if type or action is unknown."""
	return dict([(helpers[x]['type'], x) 
	             for x in helpers.keys() 
				 if helpers[x]['action'] == action]
				)[type]

def make_args(helper, input, output, args=''):
	"""Substitute for command line args"""
	esc = {'z': '%', 'i': "###input###", 'o': "###output###", 'a': args}
	out = ""
	fmt = helpers[helper]['cmd']
	fmt = fmt.replace('%%', '%z')
	fmt = fmt.split('%')
	while fmt:
		out += fmt[0]
		if len(fmt) <= 1:
			break
		fmt = fmt[1:]
		code = fmt[0] and fmt[0][0]
		if code in esc.keys():
			fmt[0] = esc[code] + fmt[0][1:]
		else:
			raise conf.ErrorBadFormat
	out = out.split()
	if "###input###" in out:
		out[out.index("###input###")] = input
	if "###output###" in out:
		out[out.index("###output###")] = output
	return out

def est_decoded_size(filename):
	# This could be much better
	tp = conf.getfiletype(filename)
	factors = {'mp3': 16.5, 'ogg': 16.5, 'flac': 2.8, 'wav': 1}
	return os.stat(filename).st_size * factors[tp]
