### #!/usr/bin/python

#  - main.py - 
# This file is part of mp3togo

# Convert audio files to play on a mp3 player
#
# (c) Simeon Veldstra 2004, 2006  <reallifesim@gmail.com>
#
# This software is free.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You may redistribute this program under the terms of the
# GNU General Public Licence version 2
# Available in this package or at http://www.fsf.org


"""Convert a list of files.

	Find a list of audio files in various 
	formats and qualities and convert them
	to a uniform format suitable for a 
	portable player."""

import sys
import os
import time
import termios
import fcntl

import mp3togo
import mp3togo.filelist as filelist
import mp3togo.conf as conf
import mp3togo.track as track
import mp3togo.task as task
import mp3togo.pool as pool
import mp3togo.cache as cache
import mp3togo.cluster as cluster


def fail(mesg='', code=1):
	if mesg:
		print >>sys.stderr, mesg
	sys.exit(code)

def main(argv):

	# Read in configuration
	try:
		opts = conf.Options(argv)
	except mp3togo.options.Fail, msg:
		print conf.Options().usage()
		fail(str(msg))

	# Cluster mode is handled in the cluster module
	if opts['clusterslave']:
		cluster.slave(opts)
		return 0

	# Are we creating a cache?
	if opts['makecache']:
		if not opts['cachesize']:
			opts.log(1, "Must specify a --cache-size to create a cache")
			fail()
		if cache.create(opts['makecache'], opts):
			opts.log(1, "Cache successfully created at %s" % opts['makecache'])
			return 0
		else:
			opts.log(1, "Create cache at %s failed" % opts['makecache'])
			fail()
	
	# Are we using a cache?
	if opts['usecache']:
		try:
			cooked = cache.Cache(opts['usecache'], opts)
		except:
			opts.log(1, "Error reading cache at %s" % opts['usecache'])
	else:
		cooked = None
		
	# Check for sanity
	if not opts.bin[opts['encoder']]:
		opts.log(1, "Encoder binary '%s' not found! Quiting." % opts['encoder'])
		fail('')
	if not opts.bin['normalize-audio'] and not opts['nonormal']:
		opts.log(1, "'normalize-audio' binary not found. Normalization disabled.")
		opts['nonormal'] = True
	if not os.path.isdir(opts['playerdir']):
		opts.log(1, "target_dir does not exist")
		fail('')

	# Build a playlist
	playlist = filelist.FileList(opts)

	# We could move these into FileList,
	# but, we may want interactive import buttons in a GUI later.
	if opts['playlist']:
		playlist.addplaylist(opts['playlist'])

	if opts['arg_files']:
		playlist.addfiles(opts['arg_files'])

	if opts['readxmms']:
		try:
			try:
				playlist.addXMMS()
			except conf.ErrorNoXMMSControl:
				opts.log(1, "python-xmms is not installed. Can't open XMMS.")
		except conf.ErrorXMMSNotRunning:
			opts.log(1, "XMMS Does not seem to be running.")

	if opts['readamarok']:
		if opts.bin['dcop']:
			try:
				playlist.addAmarok()
			except conf.ErrorAmarokNotRunning:
				opts.log(1, "Amarok Does not seem to be running.")
		else:
			opts.log(1, "dcop is not installed. Can't open Amarok.")

	# Offer a hint
	if len(playlist) == 0:
		print opts.usage()
		opts.log(1, "No files to convert.")
		fail('')

	# Start work
	if opts['cluster']:
		cluster.slavedriver(playlist, opts, cooked)
		return

	# Nothing fancy
	execute_sequential(playlist, opts, cooked)
	return


def execute_sequential(playlist, opts, cooked=None):
	"""Run the conversions one at a time"""
	print "mp3togo %s\n" % mp3togo.version
	print "<space> or 'p' to pause, <esc> or 'q' to quit\n"
	c = ''
	start_time = time.time()
	good_ones = 0
	bad_ones = 0
	playlist.run()

	try:
		# Terminal sushi
		fd = sys.stdin.fileno()
		oldterm = termios.tcgetattr(fd)
		newattr = termios.tcgetattr(fd)
		newattr[3] = newattr[3] & ~termios.ICANON & ~termios.ECHO
		termios.tcsetattr(fd, termios.TCSANOW, newattr)

		oldflags = fcntl.fcntl(fd, fcntl.F_GETFL)
		fcntl.fcntl(fd, fcntl.F_SETFL, oldflags | os.O_NONBLOCK)

		if opts['verbosity']:
			tryp = conf.try_print
		else:
			tryp = lambda x: None

		for name in playlist:
			track_start = time.time()
			tryp("(%d/%d) %s: \n" % (playlist.cur_file() + 1, len(playlist), name))
			trk = track.Track(name, opts, cooked)
			pl = pool.Pool(opts)    # make a new one for every track? Huh?
			if pl.add_track(trk):
				tsk = trk.getfirst()
				try:
					while tsk:
						tsk.run()
						if tsk.name in ('Start', 'Done', 'Cleaning'):
							tsk.wait()
							tsk = tsk.next()
							continue
						while tsk.status() == task.RUNNING:
							#parse output
							tryp("\r")
							pcent = tsk.output()
							if 0 < pcent < 101:
								bar = '#' * int(pcent/5.0) + ('_' * (20 - int(pcent/5.0)))
							else:
								bar = ""
								pcent = 0
							tryp("%s: %3.d%% %s" % (tsk.name, int(pcent), bar))
							# Get keyboard input
							time.sleep(0.01)
							for busy in range(5):
								try:
									c = sys.stdin.read(1)
									break
								except IOError: pass
							if c in ('q', 'Q', chr(27)):
								raise KeyboardInterrupt
							elif c in (' ', 'p', 'P'):
								if tsk.pause():
									tryp("\r  [PAUSED] hit <enter> to resume       ")
									while 1:
										try:
											c = sys.stdin.read(1)
											tsk.run()
											break
										except IOError: pass
						tm = tsk.elapsed_time()
						ts = format_time(tm)
						tsk.wait()
						tryp("\r                                                 ")
						if tsk.status() == task.DONE:
							if ts:
								tryp("\r%s: Done. %s elapsed.\n" % (tsk.name, ts))
							else:
								tryp("\r%s: Done.\n" % tsk.name)
							tsk = tsk.next()
						else:
							tryp("\r%s: Failed. Cleaning up.\n" % tsk.name)
							while tsk:
								tryp("Undo: %s\n" % tsk.name)
								tsk.undo()
								tsk = tsk.prev()
				except KeyboardInterrupt:
					tsk.stop()
					tryp("Stopping...\n")
					while tsk:
						tryp("Undo: %s\n" % tsk.name)
						tsk.undo()
						tsk = tsk.prev()
					sys.exit(1)
					pass
			else:
				opts.log(1, "Out of space. Exiting.")
				break
	
			pl.rm_track(trk)
			if trk() == track.DONE:
				good_ones += 1
			else:
				bad_ones += 1
			trk.close()
			del trk

			tt = time.time() - track_start
			ts = format_time(tt)
			if ts:
				tryp("Total time this track: %s\n\n" % ts)
			else:
				tryp('\n') 
	
		tm = time.time() - start_time
		ts = format_time(tm)
		if bad_ones:
			if bad_ones == 1:
				bad_str = ", 1 track failed."
			else:
				bad_str = ", %d tracks failed." % bad_ones
		else:
			bad_str = "."
		#bytes = format_bytes(pl.produced_bytes) # you are using a new pool each time!
		tryp("\n%d tracks done%s %s total time elapsed.\n" % 
		      (good_ones, bad_str, ts))
		
	finally:
		# Fry some fish:
		termios.tcsetattr(fd, termios.TCSAFLUSH, oldterm)
		fcntl.fcntl(fd, fcntl.F_SETFL, oldflags)


def format_time(tm):
	tm = int(tm)
	if tm:
		hr, min, sec = (tm/3600, (tm%3600)/60, (tm%3600)%60)
		ts = "%ds " % sec
		if min > 0:
			ts = ("%dm " % min) + ts
		if hr > 0:
			ts = ("%dh " % hr) + ts
		return ts[:-1]
	else:
		return ''

def format_bytes(bytes):
	k = 1024.0
	m = 1024 * k
	g = 1024 * m
	if 0 < bytes < k:
		ret = str(int(bytes/k)) + " KB"
	elif k <= bytes < m:
		ret = str(int(bytes/k)) + " KB"
	elif m <= bytes < g:
		ret = str(int(bytes/m)) + " MB"
	else:
		ret = str(bytes) + " Bytes"


if __name__ == "__main__": main(sys.argv)

