#  - options.py - 
# This file is part of mp3togo

# Manage command line and config file program options
#
# (c) Simeon Veldstra 2006  <reallifesim@gmail.com>
#
# This software is free.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You may redistribute this program under the terms of the
# GNU General Public Licence version 2
# Available in this package or at http://www.fsf.org

# requires python-pyvorbis and python-id3 and lame
# and mpg321

"""Configuration manager for command line programs.

	Instantiate Options with argv as an argument and command
	line options will be read in. With no argv a default set
	of options will be initialized. If ~/.mp3togo exists, it
	will be read in before command line options are
	processed. The command line takes precedence over the
	config file unless a config file is specified on the
	command line or through the conffile argument to the
	constructor. In this case, the file is processed last
	and overrides any command line options. To totaly
	disable reading in config files, set readconf=False in
	the constructor."""



import sys
import os
import UserDict
import getopt
import threading

class Fail(Exception):
	def __init__(self, msg):
		self.msg = msg 
	
	def __str__(self):
		return self.msg + "\n"

class Options(UserDict.DictMixin):
	"""Class: mp3togo.conf.Options
The Options class obtains configuration information and
makes it available to the running program in a dictionary
interface. 
"""
	def __init__(self, config_file):
		# The lock
		self._mutex = threading.Lock()

		# The name of the program
		self.name = ''
		self.version = ''

		# The log
		self._log = []

		# The option dictionary
		self._d = {} 

		# Add program options here
		# ((name, short option, long option, default value, hook, help text), ...)
		self._set = [
 ('conffile', 'c', 'config', config_file, self._conffile, 
 '''The location of the config file. (If it exists.)'''),
 ('saveconffile', 's', 'save-config', False, self._saveconffile, 
 '''Write out a new config file with the options given and exit. To save the configuration to a different file, set --config to the desired filename. The file, if it exists, will be overwritten.'''),
 ('noconffile', '', 'no-config-file', False, self._noconffile, 
 '''Disable reading from and writing to the config file.'''),
 ('verbosity', 'v', 'verbose', 1, self._verbosity, 
 '''Set the verbosity level. A value of 0 will produce no output (except option parsing errors).'''),
 ('help', 'h', 'help', False, self._help, 
 '''Print this message.'''),
 ('arg_files', '', '', [], self._arg_files, '')]

		# Binaries to check for:
		self.bin = {}
		
		# Third party modules to check for:
		self.mod = {}

		self._help_preamble = ""
		
		# Don't save these options to the config file:
		self._nosave_options =  ['saveconffile', 'conffile', 'arg_files', 'help']


	def _lock(self):
		self._mutex.acquire()

	def _unlock(self):
		self._mutex.release()

	def _locked(self):
		return self._mutex.locked()

	def getconf(self, argv=None, conffile=None, readconf=True):
		self._lock()

		#Do any pre configuration chores
		try:
			self._pre_hook()
		except:
			self._unlock()
			raise
		self._readconf = readconf

		#Default values:
		self._unlock()
		for i in self._set:
			self[i[0]] = i[3] # hooks run
		self._lock()
		
		#Read comand line options
		olist, loose = {}, []
		if argv:
			self.name = os.path.basename(argv[0])
			argv = argv[1:]
			sopts = ''
			lopts = []
			for i in self._set:
				if i[1]:
					sopts += i[1]
					if i[3] is not False:
						sopts += ':'
				if i[2]:
					if i[3] is not False:
						lopts.append(i[2] + '=')
					else:
						lopts.append(i[2])
			try:
				olist, loose = getopt.getopt(argv, sopts, lopts)
			except getopt.GetoptError:
				self._unlock()
				raise Fail, "Bad arguments."
			if '--no-config-file' in olist:
				self._readconf = False
		
		#Read in config file
		if conffile:
			self._d['conffile'] = conffile
		oldconf = self._d['conffile']
		self._unlock()
		self.load_config_file()
		self._lock()

		#Sort out loose files from the command line
		if loose:
			files = []
			for f in loose:
				absf = os.path.expanduser(f)
				absf = os.path.abspath(absf)
				if os.path.exists(absf):
					files.append(absf)
			self._d['arg_files'] = files

		#Process options.
		self._unlock() # may execute a hook function
		if olist:
			for opt, val in olist:
				opt = opt.strip('-')
				val = val.strip()
				for line in self._set:
					if opt == line[1] or opt == line[2]:
						opttype = type(line[3])
						try:
							if type(val) == opttype:
								self[line[0]] = val
							elif opttype == type(1):
								self[line[0]] = int(val)
							elif opttype == type(1.0):
								self[line[0]] = float(val)
							elif opttype == type(False):
								self[line[0]] = True
						except ValueError:
							raise Fail, "Expecting Numeric argument!"
						break

		#Offer help, if asked:
		#unlocked
		if self.opt('help'):
			print self.usage()
			sys.exit(0)

		#Save the conf file and exit if so instructed:
		#unlocked
		if self.opt('saveconffile'):
			self.log(1, "Saving config to '%s'" % self._d['conffile'])
			self.save_config_file()
			self.log(1, "Exiting.")
			sys.exit(0)

		#Read in new config file if a new one has been specified:
		#unlocked
		if self.opt('conffile') != oldconf:
			self.load_config_file()

		#Check for binaries:
		self._lock()
		search = os.environ['PATH'].split(':')
		for bin in self.bin.keys():
			if not self.bin[bin]:
				for spot in search:
					if os.path.exists(os.path.join(spot, bin)):
						self.bin[bin] = os.path.join(spot, bin)
						break
				if self.bin[bin]:
					self.log(4, "%s binary found at %s" % (bin, self.bin[bin]))
				else:
					self.log(2, "%s binary not found" % bin)

		#Check for third party modules:
		for name in self.mod.keys():
			try:
				if '.' in name:
					__import__(name, [], [], [name.split('.')[1]])
					self.mod[name] = True
				else:
					__import__(name)
			except ImportError:
				msg = "Third party module: " + name
				msg += "is not available. Doing without"
				self.log(2, "Third party module: %s is not available. Doing without" % name)
				self.mod[name] = False
			else:
				self.log(4, "Third party module: %s found" % name)
				self.mod[name] = True
				
		#Do any post configuration chores
		self._post_hook()
		self._unlock()

	def save_config_file(self):
		if self.opt('conffile') and self._readconf:
			self._lock()
			try:
				fp = file(self._d['conffile'], 'w')
				for key in self._d.keys():
					if key in self._nosave_options:
						continue
					fp.write("%s=%s\n" % (key, self._d[key]))
				fp.close()
			except IOError:
				self.log(1, "Failed to save config file")
			self._unlock()

	def load_config_file(self):
		"""Load the config file specified by the 'conffile' option, if it exists."""
		# All state access through __getitem__ and __setitem__
		# Don't lock, except to directly manipulate state
		if self.opt('conffile') and \
		  os.path.exists(os.path.expanduser(self['conffile'])) and self._readconf:
			filename = os.path.expanduser(self['conffile'])
			try:
				fp = file(filename, 'r')
				lines = fp.readlines()
				fp.close()
				conf = {}
				for line in lines:
					line = line[:-1]
					line = line.split('=', 1)
					if len(line) == 2:
						conf[line[0].lower()] = line[1].strip()
				for key in conf.keys():
					if self.has_key(key):
						if type(self[key]) == type(conf[key]):
							self[key] = conf[key]
						elif type(self[key]) == type(1):
							self[key] = int(conf[key])
						elif type(self[key]) == type(1L):
							self[key] = long(conf[key])
						elif type(self[key]) == type(1.0):
							self[key] = float(conf[key])
						elif type(self[key]) == type(True):
							if conf[key].lower() in ('no', 'off', 'false', 'disable', '0') or\
							  not conf[key]:
								self[key] = False
							else:
								self[key] = True
					else:
						self.log(2, "Unknown entry in config file: %s" % key)
			except:
				self.log(1, "Error reading config file")

	def opt(self, name):
		self._lock()
		if self._d.has_key(name):
			out = self._d[name]
		else:
			out = None 
		self._unlock()
		return out

	def __getitem__(self, name):
		self._lock()
		if self._d.has_key(name):
			self._unlock()
			return self._d[name]
		else:
			self._unlock()
			raise KeyError 

	def __setitem__(self, name, value):
		"""Set value if key exists and type is correct."""
		for i in self._set:
			if i[0] == name:
				if type(i[3]) != type(value):
					raise ValueError
				if i[4]:
					try:
						self._lock()
						apply(i[4], [name, value])
					finally:
						self._unlock()
				else:
					self._lock()
					self._d[name] = value
					self._unlock()
				return
		raise IndexError

	def keys(self):
		self._lock()
		out = self._d.keys()
		self._unlock()
		return out
	
	def __delitem__(self, item):
		"""conf items shouldn't be deleted. Object methods shouldn't
		raise an exception when called. Item deletion will fail fairly 
		silently."""
		self.log(4, "Warning, attempt to delete conf item was foiled.")

	def log(self, level, msg):
		"""Print a log message. Level must be at least 1. If the level
		is below the current verbosity level, nothing happens."""
		self._log.append((level, msg))
		# Don't lock
		if level < 1:
			raise Fail, "log() called with insufficiently high level."
		if level <= self._d['verbosity']: 
			try:
				print >>sys.stderr, msg
			except:
				pass

	def usage(self):
		"""Returns a pretty printed usage message with the available
		command line options in a string."""
		msg = "%s version %s\n" % (self.name, self.version)
		msg += self._help_preamble
		msg += "\nOptions:\n"
		for option in self._set:
			if option[1] or option[2]: 
				if option[3] is not False:
					if option [1]:
						line = "  -%s %s" % (option[1], option[3])
						if option[2]:
							line += ", "
					if option[2]:
						line += "  --%s %s" % (option[2], option[3])
				else:
					if option[1]:
						line = "  -%s" % option[1]
					else:
						line = ""
					if option[1] and option[2]:
						line += ",  "
					if option[2]:
						line += "  --%s" % option[2]
				msg += line + "\n"
				line = ""
				if option[5]:
					desc = option[5].replace('<br>', ' <br> ')
					desc = desc.replace('<li>', ' <li> ')
					desc = desc.replace('\n', ' ').split()
					while len(desc):
						line = "    "
						if desc[0] == '<li>':
							line += "  "
							del desc[0]
						if desc[0] == '<br>':
							del desc[0]
						while 1:
							if len(line) + len(desc[0]) > 60 or desc[0] in ('<br>', '<li>'):
								msg += line + "\n"
								line = ""
								break
							else:
								line += " " + desc[0] 
								del desc[0]
								if len(desc) == 0:
									msg += line + "\n\n"
									line = ""
									break
		return msg

	def manpage(self):
		"""Like usage, but formatted as a docbook sgml fragment for inclusion
		in a Debian man page.""" # Cheating!
		msg = "\n<variablelist>\n"
		for option in self._set:
			if option[1] or option[2]:
				msg += "  <varlistentry>\n"
				msg += "    <term>\n"
				if option[1]:
					msg += "      <option>-%s</option>\n" % option[1]
				if option[2]:
					msg += "      <option>--%s</option>\n" % option[2]
				msg += "    </term>\n"
				msg += "    <listitem>\n"
				if option[5]:
					desc = option[5].replace('\n', ' ').split()
					line = "      <para>"
					while len(desc):
						while 1:
							if len(line) + len(desc[0]) > 70:
								break
							else:
								line += " " + desc[0]
								del desc[0]
								if len(desc) == 0:
									break
						msg += line + '\n'
						line = "        "
					msg += "      </para>\n"
				msg += "    </listitem>\n"
				msg += "  </varlistentry>\n"
		msg += "</variablelist>\n\n"
		return msg



				

	# These are callback stubs to be overridden in subclasses:
	def _post_hook(self):
		pass

	def _pre_hook(self):
		pass

	def _arg_files(self, name, value):
		self._d['arg_files'] = value

	def _conffile(self, name, value):
		self._d['conffile'] = value

	def _saveconffile(self, name, value):
		self._d['saveconffile'] = value

	def _noconffile(self, name, value):
		self._d['noconffile'] = value

	def _verbosity(self, name, value):
		self._d['verbosity'] = value

	def _help(self, name, value):
		self._d['help'] = value

