#  - pool.py - 
# This file is part of mp3togo

# Convert audio files to play on a mp3 player
# Manage disk space for program output.
#
# (c) Simeon Veldstra 2006  <reallifesim@gmail.com>
#
# This software is free.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You may redistribute this program under the terms of the
# GNU General Public Licence version 2
# Available in this package or at http://www.fsf.org


import os
import sys
import threading

import mp3togo.conf as setup
import mp3togo.track as track


SIZE = 0
USED = 1
NEED = 2


class Pool:
	"""Manage storage"""

	def __init__(self, opts):
		self._active = []
		self._tmp = [None, 0, 0]
		self._out = [None, 0, 0]

		td = opts['tempdir']
		free_tmp = os.statvfs(td).f_bfree * os.statvfs(td).f_frsize
		if opts['maxtempsize'] and free_tmp > opts['maxtempsize']:
			self._tmp[SIZE] = opts['maxtempsize']
		else:
			self._tmp[SIZE] = free_tmp
			
		pd = opts['playerdir']
		free_out = os.statvfs(pd).f_bfree * os.statvfs(pd).f_frsize
		if opts['maxsize'] and free_out > opts['maxsize']:
			self._out[SIZE] = opts['maxsize']
		else:
			self._out[SIZE] = free_out

		self.lock = threading.RLock() 
		self.produced_bytes = 0

	def add_track(self, trk):
		self.lock.acquire()
		for tsk in trk._queue:
			if not self._add_task(tsk):
				self.rm_track(trk)
				self.lock.release()
				return False
		self.lock.release()
		return True

	def _add_task(self, tsk):
		self.lock.acquire()
		tmp = self._tmp
		out = self._out
		if tsk.tmpsize < tmp[SIZE] - (tmp[USED] + tmp[NEED]) and \
		   tsk.outsize < out[SIZE] - (out[USED] + out[NEED]):
			if tsk.tmpsize > 0:
				tmp[NEED] += tsk.tmpsize
			if tsk.outsize:
				out[NEED] += tsk.outsize
			self._active.append(tsk)
			self.lock.release()
			return True
		else:
			self.lock.release()
			return False

	def rm_track(self, trk):
		self.lock.acquire()
		tdone = 0
		for tsk in trk._queue:
			self._rm_task(tsk)
			tdone += tsk.outsize
		if trk.bytes and trk() == track.DONE:
			self.produced_bytes += trk.bytes
			self._out[USED] = self._out[USED] - tdone + trk.bytes
		self.lock.release()

	def _rm_task(self, tsk):
		if tsk in self._active:
			if tsk.tmpsize:
				if tsk.tmpsize > 0:
					self._tmp[NEED] -= tsk.tmpsize
				if tsk._status == 'done':
					self._tmp[USED] += tsk.tmpsize
			if tsk.outsize:
				self._out[NEED] -= tsk.outsize
				if tsk._status == 'done':
					self._out[USED] += tsk.outsize
			del self._active[self._active.index(tsk)]

	def tmp_willgrow(self):
		self.lock.acquire()
		ret = 0
		for tsk in self._active:
			if tsk.tmpsize < 0:
				ret += -tsk.tmpsize
		self.lock.release()
		return ret

	def tmpspace(self):
		self.lock.acquire()
		tmp = self._tmp[SIZE] - (self._tmp[USED] + self._tmp[NEED])
		self.lock.release()
		return tmp

	def outspace(self):
		self.lock.acquire()
		out = self._out[SIZE] - (self._out[USED] + self._out[NEED])
		self.lock.release()
		return out

	def run(self):
		"""Keep an eye on the situation."""
		# Will need this later
		pass

	def normalize(self):
		"""Recalculate the available space."""
		self.lock.acquire()
		pass
		self.lock.release()



